;; Generated package description from m-buffer.el  -*- no-byte-compile: t -*-
(define-package "m-buffer" "0.16.1.0.20241215.171432" "List-Oriented, Functional Buffer Manipulation" '((seq "2.14")) :commit "5e7714835b2289f61dad24c0b5cf98d28fc313b0" :url "https://elpa.gnu.org/packages/m-buffer.html" :authors '(("Phillip Lord" . "phillip.lord@russet.org.uk")) :maintainer '("Phillip Lord" . "phillip.lord@russet.rg.uk"))
